#!/usr/bin/env python3

# tuner.py - LSDJ tuner v1.0
#
# Written in 2019 by equa
#
# see `./tuner.py -h` for help.
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along
# with this software. If not, see <http://creativecommons.org/publicdomain/zero/1.0/>.

import argparse
import math
import sys

MIDDLE_C = 220 * 2 ** (1 / 4)
NOTE_COUNT = 108

def reg2freq(r):
    return 4194304 / (4 * 8 * (2048 - r))

def freq2reg(f):
    r = round(2048 - 131072 / f)
    if r <= 0:
        r = 0
        print(f'frequency {f} too low')
    elif r >= 2048:
        r = 2047
        print(f'frequency {f} too high')

    return r

def gen_et_octave(n):
    return [2 ** (x / n) for x in range(n)]

# LSDj does not have lowercase, so we replace flat signs for now
def fix_name(s):
    return s.replace('b', '*')

def gen_tuning_table(scale, center, names):
    (name, freq) = center
    def get_note(offset):
        f = 2 ** (offset // len(scale)) * scale[offset % len(scale)] * freq
        n = '{:<2}{:X}'.format(
            names[(offset + names.index(name)) % len(scale)],
            math.floor(math.log2(f) - math.log2(MIDDLE_C)) + 4
        )
        return (n, f)

    x = 0
    while get_note(x)[1] > 64:
        x -= 1
    x += 1

    return [get_note(n) for n in range(x, x + NOTE_COUNT)]

def cents_diff(x, y):
    return round((math.log2(y) - math.log2(x)) * 1200)

# Display available characters and stuff
def display_tunes(table):
    print('Name     Freq Hex     Result Off')
    for (name, freq) in table:
        print('{:3} {:9.2f} {:04X} {:9.2f} {:3d}'.format(
            name,
            freq,
            freq2reg(freq),
            reg2freq(freq2reg(freq)),
            cents_diff(freq, reg2freq(freq2reg(freq)))
        ))

def edit_file(rom, table, octave):
    def replace_byte_in(ba, find, offset, byte, count):
        if ba.count(find) != count:
            print(f'could not replace octave switch code {find}:(')
            sys.exit(1)

        return ba.replace(find,
                          find[0:offset] + bytes([byte]) + find[offset + 1:])

    ba = bytearray(rom)
    names_index = rom.find(b'C 2\x00C#2\x00')
    freqs_index = rom.find(b'\x2c\x00\x9d\x00')

    if names_index < 0 or freqs_index < 0:
        print('could not find name/frequency table! :(')
        sys.exit(1)

    for (name, freq) in table:
        ba[names_index:names_index + 4] = bytes(fix_name(name), 'utf-8') + b'\x00'
        reg = freq2reg(freq)
        ba[freqs_index:freqs_index + 2] = bytes([reg % 256, reg >> 8])
        names_index += 4
        freqs_index += 2

    # replace octaves in transposition
    ba = replace_byte_in(ba, b'\x7b\xc6\x0c\xf8', 2, octave, 1)
    ba = replace_byte_in(ba, b'\x7b\xc6\xf4\xf8', 2, 256 - octave, 1)

    ba = replace_byte_in(ba, b'\x79\xc6\x0c\xf8\x09', 2, octave, 1)
    ba = replace_byte_in(ba, b'\x79\xc6\xf4\xf8\x09', 2, 256 - octave, 1)

    ba = replace_byte_in(ba, b'\x0e\x0c\xf8\x09', 1, octave, 2)

    return bytes(ba)

def main():
    parser = argparse.ArgumentParser(
        description = 'Adjust the LSDj tuning tables.',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog =
        '''Will generate equal temperament based on the number of note names given.

Example:
    ./tuner.py lsdj.gb lsdj-19tet.gb C C# Db D D# Eb E E# F F# Gb G G# Ab A A# Bb B B#'''
    )

    parser.add_argument('in_rom', metavar='INPUT',
                        type=argparse.FileType('rb'),
                        help='Vanilla LSDj input ROM')

    parser.add_argument('out_rom', metavar='OUTPUT',
                        type=argparse.FileType('wb'),
                        help='ROM output path')

    parser.add_argument('scale', metavar='NOTE',
                        type=str,
                        nargs='+',
                        help='Scale note name (starting at middle C)')

    args = parser.parse_args()

    table = gen_tuning_table(
        # Inspect and change this to use a non-ET scale!
        gen_et_octave(len(args.scale)),
        (args.scale[0], MIDDLE_C),
        args.scale
    )

    rom = args.in_rom.read()
    args.out_rom.write(edit_file(rom, table, len(args.scale)))
    print('ROM successfully created.')

    display_tunes(table)

main()
